@@ -16,35 +16,51 @@ struct ShapeID {
   std::string font;
   unsigned int index;
   double size;
+  double tracking;
 
-  inline ShapeID() : string(""), font(""), index(0), size(0.0) {}
-  inline ShapeID(std::string _string, std::string _font, unsigned int _index, double _size) :
+  inline ShapeID() : string(""), font(""), index(0), size(0.0), tracking(0.0) {}
+  inline ShapeID(std::string _string, std::string _font, unsigned int _index, double _size, double _tracking) :
     string(_string),
     font(_font),
     index(_index),
-    size(_size) {}
+    size(_size),
+    tracking(_tracking) {}
   inline ShapeID(const ShapeID& shape) :
     string(shape.string),
     font(shape.font),
     index(shape.index),
-    size(shape.size) {}
+    size(shape.size),
+    tracking(shape.tracking) {}
 
   inline bool operator==(const ShapeID &other) const {
     return (index == other.index &&
             size == other.size &&
             string == other.string &&
-            font == other.font);
+            font == other.font) &&
+            tracking == other.tracking;
   }
 };
 struct ShapeInfo {
   std::vector<unsigned int> glyph_id;
-  std::vector<int32_t> x_pos;
+  std::vector<unsigned int> glyph_cluster;
+  std::vector<int32_t> x_advance;
+  std::vector<int32_t> y_advance;
+  std::vector<int32_t> x_offset;
+  std::vector<int32_t> y_offset;
+  std::vector<int32_t> x_bear;
+  std::vector<int32_t> y_bear;
+  std::vector<int32_t> width;
+  std::vector<int32_t> height;
+  std::vector<int32_t> ascenders;
+  std::vector<int32_t> descenders;
+  std::vector<bool> may_break;
+  std::vector<bool> must_break;
+  std::vector<bool> may_stretch;
   std::vector<unsigned int> font;
   std::vector<FontSettings> fallbacks;
+  std::vector<double> fallback_size;
   std::vector<double> fallback_scaling;
-  int32_t width;
-  int32_t left_bearing;
-  int32_t right_bearing;
+  bool ltr;
 };
 namespace std {
 template <>
@@ -53,7 +69,8 @@ struct hash<ShapeID> {
     return std::hash<std::string>()(x.string) ^
       std::hash<std::string>()(x.font) ^
       std::hash<unsigned int>()(x.index) ^
-      std::hash<double>()(x.size);
+      std::hash<double>()(x.size) ^
+      std::hash<double>()(x.tracking);
   }
 };
 }
@@ -61,6 +78,19 @@ struct hash<ShapeID> {
 class HarfBuzzShaper {
 public:
   HarfBuzzShaper() :
+  // Public
+  glyph_id(),
+  glyph_cluster(),
+  fontfile(),
+  fontindex(),
+  fontsize(),
+  string_id(),
+  x_pos(),
+  y_pos(),
+  advance(),
+  ascender(),
+  descender(),
+  must_break(),
   width(0),
   height(0),
   left_bearing(0),
@@ -72,12 +102,15 @@ public:
   pen_x(0),
   pen_y(0),
   error_code(0),
+  // Private
   cur_lineheight(0.0),
   cur_align(0),
   cur_string(0),
   cur_hjust(0.0),
   cur_vjust(0.0),
   cur_res(0.0),
+  shape_infos(),
+  may_stretch(),
   line_left_bear(),
   line_right_bear(),
   line_width(),
@@ -98,12 +131,18 @@ public:
     hb_buffer_destroy(buffer);
   };
 
-  static std::vector<unsigned int> glyph_id;
-  static std::vector<unsigned int> glyph_cluster;
-  static std::vector<unsigned int> string_id;
-  static std::vector<int32_t> x_pos;
-  static std::vector<int32_t> y_pos;
-  static std::vector<int32_t> x_mid;
+  std::vector<unsigned int> glyph_id;
+  std::vector<unsigned int> glyph_cluster;
+  std::vector<std::string> fontfile;
+  std::vector<unsigned int> fontindex;
+  std::vector<double> fontsize;
+  std::vector<unsigned int> string_id;
+  std::vector<int32_t> x_pos;
+  std::vector<int32_t> y_pos;
+  std::vector<int32_t> advance;
+  std::vector<int32_t> ascender;
+  std::vector<int32_t> descender;
+  std::vector<bool> must_break;
   int32_t width;
   int32_t height;
   int32_t left_bearing;
@@ -114,29 +153,26 @@ public:
   int32_t left_border;
   int32_t pen_x;
   int32_t pen_y;
-  static ShapeInfo last_shape_info;
 
   int error_code;
 
-  bool shape_string(const char* string, const char* fontfile, int index,
+  bool shape_string(const char* string, FontSettings& font_info,
                     double size, double res, double lineheight,
                     int align, double hjust, double vjust, double width,
                     double tracking, double ind, double hang, double before,
-                    double after);
-  bool add_string(const char* string, const char* fontfile, int index,
-                  double size, double tracking);
+                    double after, bool spacer);
+  bool add_string(const char* string, FontSettings& font_info,
+                  double size, double tracking, bool spacer);
+  bool add_spacer(FontSettings& font_info, double height, double width);
   bool finish_string();
 
-
-  bool single_line_shape(const char* string, FontSettings font_info, double size,
-                         double res);
+  ShapeInfo shape_text_run(const char* string, FontSettings& font_info, double size,
+                           double res, double tracking);
 
 private:
   static UTF_UCS utf_converter;
   static LRU_Cache<std::string, std::vector<int> > bidi_cache;
   static LRU_Cache<ShapeID, ShapeInfo> shape_cache;
-  static ShapeID last_shape_id;
-  static ShapeID temp_shape_id;
   hb_buffer_t *buffer;
   double cur_lineheight;
   int cur_align;
@@ -144,17 +180,8 @@ private:
   double cur_hjust;
   double cur_vjust;
   double cur_res;
-  double cur_tracking;
-  static std::vector<int32_t> x_advance;
-  static std::vector<int32_t> x_offset;
-  static std::vector<int32_t> left_bear;
-  static std::vector<int32_t> right_bear;
-  static std::vector<int32_t> top_extend;
-  static std::vector<int32_t> bottom_extend;
-  static std::vector<int32_t> ascenders;
-  static std::vector<int32_t> descenders;
-  static std::vector<bool> may_break;
-  static std::vector<bool> must_break;
+  std::vector<ShapeInfo> shape_infos;
+  std::vector<bool> may_stretch;
   std::vector<int32_t> line_left_bear;
   std::vector<int32_t> line_right_bear;
   std::vector<int32_t> line_width;
@@ -171,13 +198,14 @@ private:
   int32_t space_after;
 
   void reset();
-  bool shape_glyphs(hb_font_t *font, const uint32_t *string, unsigned int n_chars);
   bool shape_embedding(const uint32_t* string, unsigned start, unsigned end,
                        unsigned int string_length, double size, double res,
-                       std::vector<hb_feature_t>& features, bool emoji);
+                       double tracking, std::vector<hb_feature_t>& features,
+                       bool emoji, ShapeInfo& shape_info);
   hb_font_t* load_fallback(unsigned int font, const uint32_t* string,
                            unsigned int start, unsigned int end, int& error,
-                           double size, double res, bool& new_added);
+                           double size, double res, bool& new_added,
+                           ShapeInfo& shape_info);
   bool fallback_cluster(unsigned int font, std::vector<unsigned int>& char_font,
                         unsigned int from, unsigned int& start, unsigned int& end);
   void annotate_fallbacks(unsigned int font, unsigned int offset,
@@ -186,7 +214,11 @@ private:
                           bool& needs_fallback, bool& any_resolved, bool ltr,
                           unsigned int string_offset);
   void fill_shape_info(hb_glyph_info_t* glyph_info, hb_glyph_position_t* glyph_pos,
-                       unsigned int n_glyphs, hb_font_t* font, unsigned int font_id);
+                       unsigned int n_glyphs, hb_font_t* font, unsigned int font_id,
+                       unsigned int cluster_offset, ShapeInfo& shape_info,
+                       int32_t tracking);
+  void fill_glyph_info(const uint32_t* string, int end, ShapeInfo& shape_info);
+  size_t fill_out_width(size_t from, int32_t max, size_t shape, int& breaktype);
 
   inline double family_scaling(const char* family) {
     if (strcmp("Apple Color Emoji", family) == 0) {
@@ -196,40 +228,50 @@ private:
     }
     return 1;
   }
-  inline bool glyph_is_linebreak(int id) {
+  inline bool glyph_is_linebreak(int32_t id) {
     switch (id) {
-    case 10: return true;
-    case 11: return true;
-    case 12: return true;
-    case 13: return true;
-    case 133: return true;
-    case 8232: return true;
-    case 8233: return true;
+    case 10: return true;    // Line feed
+    case 11: return true;    // Vertical tab
+    case 12: return true;    // Form feed
+    case 13: return true;    // Cariage return
+    case 133: return true;   // Next line
+    case 8232: return true;  // Line Separator
+    case 8233: return true;  // Paragraph Separator
     }
     return false;
   }
 
-  inline bool glyph_is_breaker(int id) {
+  inline bool glyph_is_breaker(int32_t id) {
     switch (id) {
-    case 9: return true;
-    case 32: return true;
-    case 5760: return true;
-    case 6158: return true;
-    case 8192: return true;
-    case 8193: return true;
-    case 8194: return true;
-    case 8195: return true;
-    case 8196: return true;
-    case 8197: return true;
-    case 8198: return true;
-    case 8200: return true;
-    case 8201: return true;
-    case 8202: return true;
-    case 8203: return true;
-    case 8204: return true;
-    case 8205: return true;
-    case 8287: return true;
-    case 12288: return true;
+    case 9: return true;     // Horizontal tab
+    case 32: return true;    // Space
+    case 45: return true;    // Hyphen
+    case 173: return true;   // Soft hyphen
+    case 5760: return true;  // Ogham Space Mark
+    case 6158: return true;  // Mongolian Vowel Separator
+    case 8192: return true;  // En Quad
+    case 8193: return true;  // Em Quad
+    case 8194: return true;  // En Space
+    case 8195: return true;  // Em Space
+    case 8196: return true;  // Three-Per-Em Space
+    case 8197: return true;  // Four-Per-Em Space
+    case 8198: return true;  // Six-Per-Em Space
+    case 8200: return true;  // Punctuation Space
+    case 8201: return true;  // Thin Space
+    case 8202: return true;  // Hair Space
+    case 8203: return true;  // Zero Width Space
+    case 8204: return true;  // Zero Width Non-Joiner
+    case 8205: return true;  // Zero Width Joiner
+    case 8208: return true;  // Hyphen
+    case 8287: return true;  // Medium Mathematical Space
+    case 12288: return true; // Ideographic Space
+    }
+    return false;
+  }
+
+  inline bool glyph_may_stretch(int32_t id) {
+    switch (id) {
+    case 32: return true;    // Space
     }
     return false;
   }
